#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "z80em.h"

struct keytbl {
    char    *keyname;
    int	    row;
    int	    msk;
};
struct keytbl key_conv_table[] = {
    { "KANA",	0, 0x80 },
    { "GRAPH",	0, 0x40 },
    { "=",	0, 0x20 },
    { "EISU",	0, 0x10 },
//  { "",	0, 0x08 },
    { ";",	0, 0x04 },
    { ":",	0, 0x02 },
    { "CR",	0, 0x01 },

    { "Y",	1, 0x80 },
    { "Z",	1, 0x40 },
    { "@",	1, 0x20 },
    { "(",	1, 0x10 },
    { ")",	1, 0x08 },
//  { "",	1, 0x04 },
//  { "",	1, 0x02 },
//  { "",	1, 0x01 },

    { "Q",	2, 0x80 },
    { "R",	2, 0x40 },
    { "S",	2, 0x20 },
    { "T",	2, 0x10 },
    { "U",	2, 0x08 },
    { "V",	2, 0x04 },
    { "W",	2, 0x02 },
    { "X",	2, 0x01 },

    { "I",	3, 0x80 },
    { "J",	3, 0x40 },
    { "K",	3, 0x20 },
    { "L",	3, 0x10 },
    { "M",	3, 0x08 },
    { "N",	3, 0x04 },
    { "O",	3, 0x02 },
    { "P",	3, 0x01 },

    { "A",	4, 0x80 },
    { "B",	4, 0x40 },
    { "C",	4, 0x20 },
    { "D",	4, 0x10 },
    { "E",	4, 0x08 },
    { "F",	4, 0x04 },
    { "G",	4, 0x02 },
    { "H",	4, 0x01 },

    { "1",	5, 0x80 },
    { "2",	5, 0x40 },
    { "3",	5, 0x20 },
    { "4",	5, 0x10 },
    { "5",	5, 0x08 },
    { "6",	5, 0x04 },
    { "7",	5, 0x02 },
    { "8",	5, 0x01 },

    { "*",	6, 0x80 },
    { "+",	6, 0x40 },
    { "-",	6, 0x20 },
    { "SPACE",	6, 0x10 },
    { "0",	6, 0x08 },
    { "9",	6, 0x04 },
    { ",",	6, 0x02 },
    { ".",	6, 0x01 },

    { "INS",	7, 0x80 },
    { "DEL",	7, 0x40 },
//  { "",	7, 0x20 },
//  { "",	7, 0x10 },
//  { "",	7, 0x08 },
//  { "",	7, 0x04 },
    { "?",	7, 0x02 },
    { "/",	7, 0x01 },

    { "BREAK",	8, 0x80 },
    { "CTRL",	8, 0x40 },
//  { "",	8, 0x20 },
//  { "",	8, 0x10 },
//  { "",	8, 0x08 },
//  { "",	8, 0x04 },
//  { "",	8, 0x02 },
    { "SHIFT",	8, 0x01 },

    { "F1",	9, 0x80 },
    { "F2",	9, 0x40 },
    { "F3",	9, 0x20 },
    { "F4",	9, 0x10 },
    { "F5",	9, 0x08 },
//  { "",	9, 0x04 },
//  { "",	9, 0x02 },
//  { "",	9, 0x01 },

    { 0, 0, 0 }
};

struct titleinfo {
    char    titlename[16];
    char    tapeimg[16][256];
    int	    tapeimg_num;
    int	    offset;
    int	    key_row[6];
    int	    key_msk[6];
};
#define BUTTON_A	0
#define BUTTON_B	1
#define BUTTON_L	2
#define BUTTON_R	3
#define BUTTON_START	4


struct titleinfo ti_array[16];

void init_titleinfo(void) {
    int i, j;
    for (i=0; i<16; i++) {
	for (j=0; j<16; j++) ti_array[i].titlename[j] = 0;
	for (j=0; j<16; j++) {
	    ti_array[i].tapeimg[j][0] = 0;
	}
	for (j=0; j<6; j++) {
	    ti_array[i].key_row[j] = 0;
	    ti_array[i].key_msk[j] = 0;
	}
	ti_array[i].tapeimg_num = 0;
	/* default key config */
	ti_array[i].key_row[BUTTON_A] = 6;		/* [SPACE] */
	ti_array[i].key_msk[BUTTON_A] = 0x10;
	ti_array[i].key_row[BUTTON_B] = 6;		/* [SPACE] */
	ti_array[i].key_msk[BUTTON_B] = 0x10;
    }
}

int str2mzkeymap(char *str, int *row, int *msk) {
    int i;
    for (i=0; key_conv_table[i].keyname != 0; i++) {
	if (!strcmpi(str, key_conv_table[i].keyname)) {
	    *row = key_conv_table[i].row;
	    *msk = key_conv_table[i].msk;
	    return 1;
	}
    }
    *row = 0;
    *msk = 0;
    return 0;
}

char *get_config_str(char *str, int *len) {
    char *p;
    while (isspace(*str)) str++;
    if (*str++ != ':') {
	*len = 0;
	return NULL;
    }
    while (isspace(*str)) str++;
    if (*str == 0) {
	*len = 0;
	return NULL;
    }
    p = str;
    while (*p) p++;
    while (isspace(*(p-1))) *(--p) = 0;
    *len = p - str;
    return str;
}

int get_filesize(char *fn) {
    FILE *fp;
    int siz;
    fp = fopen(fn, "rb");
    if (fp == NULL) return 0;
    fseek(fp, 0, SEEK_END);
    siz = ftell(fp);
    fclose(fp);
    return siz;
}

void write_whole_file(FILE *fpw, char *fn) {
    FILE *fp;
    unsigned char *p;
    int siz;
    siz = get_filesize(fn);
    p = malloc(siz);
    fp = fopen(fn, "rb");
    fread(p, 1, siz, fp);
    fwrite(p, 1, siz, fpw);
    free(p);
    fclose(fp);
}

int main(int argc, char **argv) {
    FILE *fpr, *fpw;
    char buf[256];
    unsigned char tmp[32];
    char *cfgfn = "MZ700GBA.CFG";
    char corefn[256];
    char iplfn[256];
    char *p;
    int i, j, r, m;
    int line = 0;
    int title_index = -1;
    long offset = 0, size;

    if (argc >= 2) cfgfn = argv[1];

    fpr = fopen(cfgfn, "rb");
    if (fpr == NULL) {
	fprintf(stderr, "Config file '%s' was not found\n", cfgfn);
	return -1;
    }
    iplfn[0] = 0;
    strcpy(corefn, "MZ700GBA.BIN");
    while (fgets(buf, 256, fpr) != NULL) {
	line++;
	if (buf[0] == '#') {
	    continue;
	} else if (!strncmpi(buf, "CORE", 4)) {
	    p = get_config_str(buf+4, &i);
	    if (p==NULL) { fprintf(stderr, "%d: Parameter missing\n",line); fclose(fpr); return -1; }
	    strncpy(corefn, p, 256);
	    if (get_filesize(p) == 0) {
		fprintf(stderr, "%d: Cannot find '%s'\n", line, p);
		fclose(fpr);
		return -1;
	    }

	} else if (!strncmpi(buf, "IPLROM", 6)) {
	    p = get_config_str(buf+6, &i);
	    if (p==NULL) { fprintf(stderr, "%d: Parameter missing\n",line); fclose(fpr); return -1; }
	    strncpy(iplfn, p, 256);
	    if (get_filesize(p) == 0) {
		fprintf(stderr, "%d: Cannot find '%s'\n", line, p);
		fclose(fpr);
		return -1;
	    }

	} else if (!strncmpi(buf, "TITLE", 5)) {
	    if (title_index >= 0 && ti_array[title_index].tapeimg_num == 0) {
		fprintf(stderr, "%d: No tape image file was specified\n", line);
		fclose(fpr);
		return -1;
	    }
	    title_index++;
	    ti_array[title_index].offset = offset;
	    if (title_index >= 16) {
		fprintf(stderr, "Too many titles\n");
		fclose(fpr);
		return -1;
	    }
	    p = get_config_str(buf+5, &i);
	    if (p==NULL) { fprintf(stderr, "%d: Parameter missing\n",line); fclose(fpr); return -1; }
	    strncpy(ti_array[title_index].titlename, p, 16);

	} else if (!strncmpi(buf, "TAPEIMG", 7)) {
	    if (title_index < 0) {
		fprintf(stderr, "%d: 'TITLE' identifier should appear first\n", line);
		fclose(fpr);
		return -1;
	    }
	    p = get_config_str(buf+7, &i);
	    if (p==NULL) { fprintf(stderr, "%d: Parameter missing\n",line); fclose(fpr); return -1; }
	    i = ti_array[title_index].tapeimg_num++;
	    strncpy(&(ti_array[title_index].tapeimg[i][0]), p, 256);
	    size = get_filesize(p);
	    if (size == 0) {
		fprintf(stderr, "%d: Cannot find '%s'\n", line, p);
		fclose(fpr);
		return -1;
	    }
	    offset += size;

	} else if (!strncmpi(buf, "A", 1)) {
	    if (title_index < 0) {
		fprintf(stderr, "%d: 'TITLE' identifier should appear first\n", line);
		fclose(fpr);
		return -1;
	    }
	    p = get_config_str(buf+1, &i);
	    if (p==NULL) { fprintf(stderr, "%d: Parameter missing\n",line); fclose(fpr); return -1; }
	    if (str2mzkeymap(p, &r, &m)) {
		ti_array[title_index].key_row[BUTTON_A] = r;
		ti_array[title_index].key_msk[BUTTON_A] = m;
	    } else {
		fprintf(stderr, "%d: Unknown key identifier: '%s'\n", line, p);
		fclose(fpr);
		return -1;
	    }

	} else if (!strncmpi(buf, "B", 1)) {
	    if (title_index < 0) {
		fprintf(stderr, "%d: 'TITLE' identifier should appear first\n", line);
		fclose(fpr);
		return -1;
	    }
	    p = get_config_str(buf+1, &i);
	    if (p==NULL) { fprintf(stderr, "%d: Parameter missing\n",line); fclose(fpr); return -1; }
	    if (str2mzkeymap(p, &r, &m)) {
		ti_array[title_index].key_row[BUTTON_B] = r;
		ti_array[title_index].key_msk[BUTTON_B] = m;
	    } else {
		fprintf(stderr, "%d: Unknown key identifier: '%s'\n", line, p);
		fclose(fpr);
		return -1;
	    }

	} else if (!strncmpi(buf, "L", 1)) {
	    if (title_index < 0) {
		fprintf(stderr, "%d: 'TITLE' identifier should appear first\n", line);
		fclose(fpr);
		return -1;
	    }
	    p = get_config_str(buf+1, &i);
	    if (p==NULL) { fprintf(stderr, "%d: Parameter missing\n",line); fclose(fpr); return -1; }
	    if (str2mzkeymap(p, &r, &m)) {
		ti_array[title_index].key_row[BUTTON_L] = r;
		ti_array[title_index].key_msk[BUTTON_L] = m;
	    } else {
		fprintf(stderr, "%d: Unknown key identifier: '%s'\n", line, p);
		fclose(fpr);
		return -1;
	    }

	} else if (!strncmpi(buf, "R", 1)) {
	    if (title_index < 0) {
		fprintf(stderr, "%d: 'TITLE' identifier should appear first\n", line);
		fclose(fpr);
		return -1;
	    }
	    p = get_config_str(buf+1, &i);
	    if (p==NULL) { fprintf(stderr, "%d: Parameter missing\n",line); fclose(fpr); return -1; }
	    if (str2mzkeymap(p, &r, &m)) {
		ti_array[title_index].key_row[BUTTON_R] = r;
		ti_array[title_index].key_msk[BUTTON_R] = m;
	    } else {
		fprintf(stderr, "%d: Unknown key identifier: '%s'\n", line, p);
		fclose(fpr);
		return -1;
	    }

	} else if (!strncmpi(buf, "START", 5)) {
	    if (title_index < 0) {
		fprintf(stderr, "%d: 'TITLE' identifier should appear first\n", line);
		fclose(fpr);
		return -1;
	    }
	    p = get_config_str(buf+5, &i);
	    if (p==NULL) { fprintf(stderr, "%d: Parameter missing\n",line); fclose(fpr); return -1; }
	    if (str2mzkeymap(p, &r, &m)) {
		ti_array[title_index].key_row[BUTTON_START] = r;
		ti_array[title_index].key_msk[BUTTON_START] = m;
	    } else {
		fprintf(stderr, "%d: Unknown key identifier: '%s'\n", line, p);
		fclose(fpr);
		return -1;
	    }
	} else {
	    p = buf;
	    while (isspace(*p)) p++;
	    if (*p == 0 || *p == '#') continue;
	    fprintf(stderr, "%d: Illegal identifier\n", line);
	    fclose(fpr);
	    return -1;
	}
    }
    fclose(fpr);
    if (title_index < 0) {
	fprintf(stderr, "%d: No title was specified\n", line);
	return -1;
    }
    if (ti_array[title_index++].tapeimg_num == 0) {
	fprintf(stderr, "%d: No tape image file was specified\n", line);
	return -1;
    }
    if (get_filesize(corefn) == 0) {
	fprintf(stderr, "Emulator core file '%s' was not found\n", corefn);
	return -1;
    }
    if (iplfn[0] == 0) {
	fprintf(stderr, "No IPL ROM image was specified\n", corefn);
	return -1;
    }

/* debug...
    for (i=0; i<title_index+1; i++) {
	fprintf(stdout, "Title: %s\n", ti_array[i].titlename);
	fprintf(stdout, "Offset: %08x\n", ti_array[i].offset);
	for (r=0; r<ti_array[i].tapeimg_num; r++)
	    fprintf(stdout, "Tape image #%d: %s\n", r, &(ti_array[i].tapeimg[r][0]));
	for (r=0; r<5; r++)
	    fprintf(stdout, "Key config #%d: Row %d, Mask %02X\n", r, 
		    ti_array[i].key_row[r], ti_array[i].key_msk[r]);
    }
*/

    /* write output */
    fpw = fopen("MZ700GBA.GBA", "wb");
    if (fpw == NULL) {
	fprintf(stderr, "Cannot create an output file\n");
	return -1;
    }
    /* emulator core */
    write_whole_file(fpw, corefn);
    /* emulator core */
    write_whole_file(fpw, iplfn);
    /* tape image index */
    for (i=0; i<title_index; i++) {
	long l;
	/* Title: 16byte */
	memcpy(&tmp[0], ti_array[i].titlename, 16);
	/* Offset: 4byte */
	l = ti_array[i].offset + (title_index+1)*32;
	tmp[16] = (unsigned char)((l      ) & 0xff);
	tmp[17] = (unsigned char)((l >>  8) & 0xff);
	tmp[18] = (unsigned char)((l >> 16) & 0xff);
	tmp[19] = (unsigned char)((l >> 24) & 0xff);
	/* Key Config Data: 10byte */
	for (j=0; j<6; j++) {
	    tmp[20+j*2] = (unsigned char)ti_array[i].key_msk[j];
	    tmp[21+j*2] = (unsigned char)ti_array[i].key_row[j] + KeyMatrix;
	}
	fwrite(tmp, 1, 32, fpw);
    }
    /* terminator */
    for (i=0; i<32; i++) tmp[i] = 0;
    offset += (title_index+1)*32;
    tmp[16] = (unsigned char)((offset      ) & 0xff);
    tmp[17] = (unsigned char)((offset >>  8) & 0xff);
    tmp[18] = (unsigned char)((offset >> 16) & 0xff);
    tmp[19] = (unsigned char)((offset >> 24) & 0xff);
    fwrite(tmp, 1, 32, fpw);
    /* tape images */
    for (i=0; i<title_index; i++) {
	for (j=0; j<ti_array[i].tapeimg_num; j++) {
	    write_whole_file(fpw, &(ti_array[i].tapeimg[j][0]));
	}
    }
    fclose(fpw);

    return 0;
}

/*
    CORE: mz700gba.bin
    IPLROM: 1Z009A.BIN
    TITLE:  NASAKE
    TAPEIMG: hubasic.mzt
    TAPEIMG: nasake.mzt
    A: SPACE
    B: SPACE
    L: 
    R: 
    START: 
 */
