#include "gba.h"

__attribute__ ((section(".iwram"), long_call)) void interrupt_setup(void);

extern const u8 cgrom[];

void init_palette(void) {
    volatile u16 *BGPalette = (u16 *)0x05000000;
    u32 i;
    for (i=0; i<16; i++) {
	*BGPalette++ = 0x0000;
	*BGPalette++ = ((i & 4) ? 0x03e0 : 0x0000) |
	               ((i & 2) ? 0x001f : 0x0000) |
		       ((i & 1) ? 0x7c00 : 0x0000);
	BGPalette += 14;
    }
}

void init_bgpattern(void) {
    volatile u32 *pd0, *pd1;
    u8 c, *ps;
    u32 i, j, d;
    ps = &cgrom[0];
    pd0 = (u32 *)0x06000000;
    pd1 = (u32 *)0x06004000;
    for (i=0; i<512; i++) {
	for (j=0; j<6; j++) {
	    c = *ps++;
	    d = 0;
	    if (c & 0x80) d |= 0x00000001;
	    if (c & 0x40) d |= 0x00000010;
	    if (c & 0x20) d |= 0x00000100;
	    if (c & 0x10) d |= 0x00001000;
	    if (c & 0x08) d |= 0x00010000;
	    if (c & 0x04) d |= 0x00100000;
	    *pd0++ = d;
	    *pd1++ = (d << 8);
	}
	*pd0++ = 0;
	*pd0++ = 0;
	*pd1++ = 0;
	*pd1++ = 0;
    }
}

void vraminit(void) {
    init_palette();
    init_bgpattern();
    *((volatile u16 *)0x04000008) = 0x1000;	/* BG0: 256x256, Map Block #16, 16 color */
    *((volatile u16 *)0x0400000a) = 0x1100;	/* BG1: 256x256, Map Block #17, 16 color */
    *((volatile u16 *)0x0400000c) = 0x1200;	/* BG2: 256x256, Map Block #18, 16 color */
    *((volatile u16 *)0x0400000e) = 0x1300;	/* BG3: 256x256, Map Block #19, 16 color */
    *((volatile u32 *)0x04000010) = 0x00000000;	/* BG0 offset, +0, +0 */
    *((volatile u32 *)0x04000014) = 0x00000004;	/* BG1 offset, +4, +0 */
    *((volatile u32 *)0x04000018) = 0x00000000;	/* BG2 offset, +0, +0 */
    *((volatile u32 *)0x0400001c) = 0x00000004;	/* BG3 offset, +4, +0 */
//    *((volatile u16 *)0x04000000) = 0x0f20;	/* Mode 0, Display BG0-3, No OBJ during HBLNK */
    *((volatile u16 *)0x04000000) = 0x0f20;	/* Mode 0, Display BG0-3, No OBJ during HBLNK */

    interrupt_setup();
}

void soundinit(void) {
    *((volatile u16 *)0x04000084) = 0x0080;	/* Sound circuit ON */
    *((volatile u32 *)0x04000080) = 0x000e6677;	/* Sound channel 2 only */
    *((volatile u16 *)0x04000068) = 0xf080;	/* Channel 2 Duty/Len/Envelope */

    *((volatile u16 *)0x04000070) = 0x0040;	/* Channel 3, Bank 1 */
    *((volatile u32 *)0x04000090) = 0xffffffff;	/* Keep high level */
    *((volatile u32 *)0x04000094) = 0xffffffff;
    *((volatile u32 *)0x04000098) = 0xffffffff;
    *((volatile u32 *)0x0400009c) = 0xffffffff;
    *((volatile u16 *)0x04000070) = 0x0000;	/* Channel 3, Bank 0 */
    *((volatile u32 *)0x04000090) = 0xffffffff;	/* Keep high level */
    *((volatile u32 *)0x04000094) = 0xffffffff;
    *((volatile u32 *)0x04000098) = 0xffffffff;
    *((volatile u32 *)0x0400009c) = 0xffffffff;
    *((volatile u32 *)0x04000070) = 0x00008000;	/* Channel 3, ON, mute */

//    *((volatile u16 *)0x04000088) = 0xc200;	/* Sampling cycle: 262.144kHz */
}

